Package: fdclassify
Title: Supervised Classification for Functional Data via Signed Depth
Version: 0.1.0
Authors@R: c(
    person("Diego Andrés", "Pérez Ruiz",
           email = "diego.perezruiz@manchester.ac.uk",
           role  = c("aut", "cre"),
           comment = c(ORCID = "0000-0002-9969-6065")),
    person("Peter", "Foster",
           role  = "ths"))
Description: Provides a suite of supervised classifiers for functional data
    based on the concept of signed depth. The core pipeline computes
    Fraiman-Muniz (FM) functional depth in either its Tukey or Simplicial
    variant, derives a signed depth by comparing each curve to a reference
    median curve via the signed distance integral, and feeds the resulting
    scalar summary into several classifiers: the k-Ranked Nearest Neighbour
    (k-RNN) rule, a moving-average smoother, a kernel-density Bayes rule,
    logistic regression on signed depth and distance to the mode, and a
    generalised additive model (GAM) classifier. Cross-validation routines
    for tuning the neighbourhood size k and parametric bootstrap confidence
    intervals are also included.
License: GPL-3
Encoding: UTF-8
Language: en-GB
RoxygenNote: 7.3.1
Depends: R (>= 4.1.0)
Imports: stats, graphics, mgcv, modeest
Suggests: testthat (>= 3.0.0), spelling, knitr, rmarkdown
Config/testthat/edition: 3
URL: https://github.com/dapr12/fdclassify
BugReports: https://github.com/dapr12/fdclassify/issues
NeedsCompilation: no
Packaged: 2026-04-22 14:53:05 UTC; mbbxkdp3
Author: Diego Andrés Pérez Ruiz [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-9969-6065>),
  Peter Foster [ths]
Maintainer: Diego Andrés Pérez Ruiz <diego.perezruiz@manchester.ac.uk>
Repository: CRAN
Date/Publication: 2026-04-23 20:10:03 UTC
Built: R 4.6.0; ; 2026-04-27 16:50:29 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(predict,krnn_fit)
S3method(print,fd_bayes_fit)
S3method(print,fd_gam_fit)
S3method(print,fd_logistic_fit)
S3method(print,fd_signed_depth)
S3method(print,krnn_cv)
S3method(print,krnn_fit)
export(bayes_depth_classify)
export(fm_depth)
export(gam_depth_classify)
export(krnn_cv)
export(krnn_fit)
export(krnn_smoother)
export(logistic_depth_classify)
export(plot_krnn_cv)
export(plot_krnn_smoother)
export(reference_curve)
export(signed_depth)
export(signed_distance_integral)
export(simulate_fd)

# ============================================================
#  krnn.R
#  k-Ranked Nearest Neighbour classifier for functional data
#  Sections 3.2 - 3.4 of Perez Ruiz (2020)
# ============================================================


# ---- Internal helpers --------------------------------------------------

#' @keywords internal
.krnn_predict_sdp <- function(sdp_new, sdp_train, y_train, k) {
  o          <- order(sdp_train)
  sdp_sorted <- sdp_train[o]
  y_sorted   <- y_train[o]
  n_tr       <- length(sdp_sorted)
  vapply(sdp_new, function(s0) {
    pos <- findInterval(s0, sdp_sorted)
    lo  <- max(1L, pos - k + 1L)
    hi  <- min(n_tr, pos + k)
    neigh_labels <- y_sorted[lo:hi]
    tab <- table(neigh_labels)
    as.integer(names(tab)[which.max(tab)])
  }, integer(1L))
}

#' @keywords internal
.smoother_prob <- function(sdp_new, sdp_train, y_train, k,
                           pre_sorted = FALSE) {
  if (!pre_sorted) {
    o         <- order(sdp_train)
    sdp_train <- sdp_train[o]
    y_train   <- y_train[o]
  }
  n_tr <- length(sdp_train)
  vapply(sdp_new, function(s0) {
    pos <- findInterval(s0, sdp_train)
    lo  <- max(1L, pos - k + 1L)
    hi  <- min(n_tr, pos + k)
    mean(y_train[lo:hi] == 0L)
  }, numeric(1L))
}

#' @keywords internal
.project_signed_depth <- function(X_new, sd_train) {
  sdi_new   <- signed_distance_integral(X_new,
                                        x_ref = sd_train$x_ref,
                                        grid  = sd_train$grid)
  depth_new <- fm_depth(X_new, grid = sd_train$grid, type = sd_train$type)
  sign(sdi_new) * depth_new
}

#' @keywords internal
.bootstrap_ci <- function(sdp_sorted, y_sorted, k, B = 200L, alpha = 0.05) {
  p_hat    <- .smoother_prob(sdp_sorted, sdp_sorted, y_sorted, k,
                             pre_sorted = TRUE)
  n        <- length(sdp_sorted)
  boot_mat <- matrix(NA_real_, nrow = n, ncol = B)
  for (b in seq_len(B)) {
    y_boot        <- stats::rbinom(n, 1L, 1 - p_hat)
    boot_mat[, b] <- .smoother_prob(sdp_sorted, sdp_sorted, y_boot, k,
                                    pre_sorted = TRUE)
  }
  se_boot <- apply(boot_mat, 1, stats::sd)
  z_boot  <- sweep(boot_mat, 1, p_hat, "-") / (se_boot + 1e-10)
  t_lo    <- apply(z_boot, 1, stats::quantile, probs = alpha / 2)
  t_hi    <- apply(z_boot, 1, stats::quantile, probs = 1 - alpha / 2)
  list(lower = pmax(0, p_hat - t_hi * se_boot),
       upper = pmin(1, p_hat - t_lo * se_boot))
}


# -----------------------------------------------------------------------
#  krnn_fit
# -----------------------------------------------------------------------

#' k-Ranked Nearest Neighbour Classifier for Functional Data
#'
#' Fits the k-RNN classifier of Perez Ruiz (2020). Curves are ranked by their
#' signed depth and each new observation is assigned to the majority class
#' among its \eqn{2k} nearest ranked neighbours (k above and k below).
#'
#' @param X_train Numeric matrix (\eqn{n_{\mathrm{train}} \times M}) of
#'   training curves (rows = curves, columns = time points).
#' @param y_train Integer vector of class labels (0/1) of length
#'   \eqn{n_{\mathrm{train}}}.
#' @param X_test Numeric matrix of test curves. If \code{NULL} (default) the
#'   training set is used for in-sample evaluation.
#' @param k Positive integer: half-neighbourhood size. The total number of
#'   neighbours per observation is \eqn{2k}. Use \code{\link{krnn_cv}} to
#'   select k.
#' @param grid Numeric vector of length \eqn{M} giving observation times.
#'   Defaults to equally spaced points on \eqn{[0, 1]}.
#' @param type Depth variant passed to \code{\link{fm_depth}}: \code{"tukey"}
#'   (default) or \code{"simplicial"}.
#' @param sd_train Optional pre-computed \code{"fd_signed_depth"} object for
#'   the training set (avoids recomputation when called repeatedly).
#'
#' @return An object of class \code{"krnn_fit"} with components:
#'   \describe{
#'     \item{predicted}{Integer vector of predicted class labels.}
#'     \item{prob}{Numeric vector of estimated probabilities
#'       \eqn{\hat{P}(y=0|\mathrm{sdp})}.}
#'     \item{sd_train}{The \code{"fd_signed_depth"} object for the training
#'       set.}
#'     \item{y_train}{The training labels.}
#'     \item{k}{The value of k used.}
#'   }
#'
#' @seealso \code{\link{krnn_cv}}, \code{\link{predict.krnn_fit}}
#'
#' @references
#' Perez Ruiz, D. A. (2020). \emph{Supervised Classification for Functional
#' Data}. PhD thesis, University of Manchester. Section 3.2.
#'
#' @examples
#' set.seed(7)
#' M <- 100; n <- 80
#' t <- seq(0, 1, length.out = M)
#' X0 <- t(replicate(n / 2, sin(2 * pi * t) + rnorm(M, sd = 0.4)))
#' X1 <- t(replicate(n / 2, cos(2 * pi * t) + rnorm(M, sd = 0.4)))
#' X  <- rbind(X0, X1)
#' y  <- rep(0:1, each = n / 2)
#' idx_tr <- sample(n, 60)
#' fit <- krnn_fit(X[idx_tr, ], y[idx_tr], X[-idx_tr, ], k = 5)
#' print(fit)
#'
#' @export
krnn_fit <- function(X_train, y_train,
                     X_test   = NULL,
                     k        = 5L,
                     grid     = NULL,
                     type     = c("tukey", "simplicial"),
                     sd_train = NULL) {
  mc      <- match.call()
  type    <- match.arg(type)
  y_train <- as.integer(y_train)
  if (!all(y_train %in% c(0L, 1L)))
    stop("'y_train' must contain only 0 and 1.")
  if (!is.matrix(X_train))
    stop("'X_train' must be a matrix.")
  k <- as.integer(k)
  if (k < 1L) stop("'k' must be a positive integer.")
  M <- ncol(X_train)
  if (is.null(grid)) grid <- seq(0, 1, length.out = M)
  if (is.null(sd_train))
    sd_train <- signed_depth(X_train, grid = grid, type = type)
  sdp_tr <- sd_train$sdp
  if (is.null(X_test)) {
    sdp_te <- sdp_tr
  } else {
    if (ncol(X_test) != M)
      stop("'X_test' must have the same number of columns as 'X_train'.")
    sdp_te <- .project_signed_depth(X_test, sd_train)
  }
  predicted <- .krnn_predict_sdp(sdp_te, sdp_tr, y_train, k)
  prob      <- .smoother_prob(sdp_te, sdp_tr, y_train, k)
  structure(
    list(predicted = predicted, prob = prob, sd_train = sd_train,
         y_train = y_train, k = k, type = type, call = mc),
    class = "krnn_fit"
  )
}

#' @param x A \code{"krnn_fit"} object.
#' @param ... Ignored.
#' @return Invisibly returns \code{x}.
#' @rdname krnn_fit
#' @export
print.krnn_fit <- function(x, ...) {
  cat("k-RNN Functional Classifier\n")
  cat("  k          :", x$k, "\n")
  cat("  Depth type :", x$type, "\n")
  cat("  Training N :", x$sd_train$N, "\n")
  invisible(x)
}


# -----------------------------------------------------------------------
#  predict.krnn_fit
# -----------------------------------------------------------------------

#' Predict Method for krnn_fit Objects
#'
#' Classifies new functional observations using a fitted
#' \code{\link{krnn_fit}} object.
#'
#' @param object A \code{"krnn_fit"} object.
#' @param newdata Numeric matrix of new curves
#'   (\eqn{n_{\mathrm{new}} \times M}).
#' @param y_true Optional integer vector of true labels for computing the
#'   misclassification rate.
#' @param ... Ignored.
#'
#' @return A list with components \code{predicted}, \code{prob}, and (if
#'   \code{y_true} is supplied) \code{misclass}.
#'
#' @examples
#' set.seed(7)
#' M <- 100; n <- 80
#' t <- seq(0, 1, length.out = M)
#' X0 <- t(replicate(n / 2, sin(2 * pi * t) + rnorm(M, sd = 0.4)))
#' X1 <- t(replicate(n / 2, cos(2 * pi * t) + rnorm(M, sd = 0.4)))
#' X  <- rbind(X0, X1)
#' y  <- rep(0:1, each = n / 2)
#' idx_tr <- sample(n, 60)
#' fit  <- krnn_fit(X[idx_tr, ], y[idx_tr], k = 5)
#' pred <- predict(fit, newdata = X[-idx_tr, ], y_true = y[-idx_tr])
#' pred$misclass
#'
#' @export
predict.krnn_fit <- function(object, newdata, y_true = NULL, ...) {
  sdp_new   <- .project_signed_depth(newdata, object$sd_train)
  predicted <- .krnn_predict_sdp(sdp_new, object$sd_train$sdp,
                                 object$y_train, object$k)
  prob      <- .smoother_prob(sdp_new, object$sd_train$sdp,
                              object$y_train, object$k)
  out <- list(predicted = predicted, prob = prob)
  if (!is.null(y_true))
    out$misclass <- mean(predicted != as.integer(y_true))
  out
}


# -----------------------------------------------------------------------
#  krnn_cv
# -----------------------------------------------------------------------

#' Cross-Validation for k-RNN Neighbourhood Size
#'
#' Selects the optimal half-neighbourhood size \eqn{k} by \eqn{R}-fold
#' cross-validation, using either the minimum-error rule or the
#' one-standard-error (1-SE) rule (Section 3.3.1 of Perez Ruiz, 2020).
#'
#' @param X Numeric matrix (\eqn{N \times M}).
#' @param y Integer vector of class labels (0/1).
#' @param k_max Maximum value of k to evaluate. Default
#'   \code{floor(nrow(X) / 4)}.
#' @param R Number of CV folds. Default \code{10}.
#' @param rule Character: \code{"min"} (default) or \code{"1se"}.
#' @param grid Numeric vector of length \eqn{M}.
#' @param type Depth variant.
#' @param seed Optional integer seed for reproducibility.
#'
#' @return An object of class \code{"krnn_cv"} with components:
#'   \describe{
#'     \item{k_opt}{Selected optimal k.}
#'     \item{cv_error}{Mean CV misclassification error for each k.}
#'     \item{cv_se}{Standard error of CV error per k.}
#'     \item{k_max}{Maximum k evaluated.}
#'     \item{R}{Number of folds used.}
#'     \item{rule}{The selection rule used.}
#'   }
#'
#' @examples
#' \donttest{
#' set.seed(3)
#' M <- 80
#' t  <- seq(0, 1, length.out = M)
#' X0 <- t(replicate(50, sin(2 * pi * t) + rnorm(M, sd = 0.4)))
#' X1 <- t(replicate(50, cos(2 * pi * t) + rnorm(M, sd = 0.4)))
#' X  <- rbind(X0, X1)
#' y  <- rep(0:1, each = 50)
#' cv <- krnn_cv(X, y, k_max = 20, R = 5, seed = 1)
#' print(cv)
#' plot_krnn_cv(cv)
#' }
#'
#' @export
krnn_cv <- function(X, y,
                    k_max = NULL,
                    R     = 10L,
                    rule  = c("min", "1se"),
                    grid  = NULL,
                    type  = c("tukey", "simplicial"),
                    seed  = NULL) {
  rule <- match.arg(rule)
  type <- match.arg(type)
  y    <- as.integer(y)
  N    <- nrow(X); M <- ncol(X)
  if (is.null(k_max)) k_max <- max(1L, floor(N / 4L))
  k_max <- as.integer(k_max); R <- as.integer(R)
  if (is.null(grid)) grid <- seq(0, 1, length.out = M)
  if (!is.null(seed)) set.seed(seed)
  folds   <- sample(rep(seq_len(R), length.out = N))
  err_mat <- matrix(NA_real_, nrow = R, ncol = k_max)
  for (r in seq_len(R)) {
    idx_te      <- which(folds == r)
    idx_tr      <- which(folds != r)
    sd_tr       <- signed_depth(X[idx_tr, , drop = FALSE],
                                grid = grid, type = type)
    sdp_te_proj <- .project_signed_depth(X[idx_te, , drop = FALSE], sd_tr)
    sdp_tr      <- sd_tr$sdp
    y_tr        <- y[idx_tr]; y_te <- y[idx_te]
    for (k in seq_len(k_max)) {
      pred          <- .krnn_predict_sdp(sdp_te_proj, sdp_tr, y_tr, k)
      err_mat[r, k] <- mean(pred != y_te)
    }
  }
  cv_error <- colMeans(err_mat)
  cv_se    <- apply(err_mat, 2, stats::sd) / sqrt(R)
  k_min    <- which.min(cv_error)
  k_opt    <- if (rule == "min") k_min else
    which(cv_error <= cv_error[k_min] + cv_se[k_min])[1L]
  structure(
    list(k_opt = k_opt, cv_error = cv_error, cv_se = cv_se,
         k_max = k_max, R = R, rule = rule),
    class = "krnn_cv"
  )
}

#' @param x A \code{"krnn_cv"} object.
#' @param ... Ignored.
#' @return Invisibly returns \code{x}.
#' @rdname krnn_cv
#' @export
print.krnn_cv <- function(x, ...) {
  cat("k-RNN Cross-Validation\n")
  cat("  Folds (R)      :", x$R,     "\n")
  cat("  Rule           :", x$rule,  "\n")
  cat("  Optimal k      :", x$k_opt, "\n")
  cat("  CV error at k* :", round(x$cv_error[x$k_opt], 4), "\n")
  invisible(x)
}


# -----------------------------------------------------------------------
#  plot_krnn_cv
# -----------------------------------------------------------------------

#' Plot a krnn_cv Object
#'
#' Plots the cross-validation error curve against the neighbourhood size k,
#' with standard-error bars and a vertical line at the selected optimum.
#'
#' @param x A \code{"krnn_cv"} object returned by \code{\link{krnn_cv}}.
#' @param ... Additional graphical parameters passed to
#'   \code{\link[graphics]{plot.default}}.
#'
#' @return Invisibly returns \code{x}.
#'
#' @examples
#' \donttest{
#' set.seed(3)
#' M <- 80
#' t  <- seq(0, 1, length.out = M)
#' X0 <- t(replicate(50, sin(2 * pi * t) + rnorm(M, sd = 0.4)))
#' X1 <- t(replicate(50, cos(2 * pi * t) + rnorm(M, sd = 0.4)))
#' X  <- rbind(X0, X1)
#' y  <- rep(0:1, each = 50)
#' cv <- krnn_cv(X, y, k_max = 20, R = 5, seed = 1)
#' plot_krnn_cv(cv)
#' }
#'
#' @export
plot_krnn_cv <- function(x, ...) {
  k_seq <- seq_len(x$k_max)
  graphics::plot(k_seq, x$cv_error,
                 type = "b", pch = 19, cex = 0.7,
                 xlab = "Number of neighbours (k)",
                 ylab = "CV misclassification error",
                 main = paste0("k-RNN cross-validation  (", x$R, "-fold)"),
                 ylim = range(c(x$cv_error - x$cv_se,
                                x$cv_error + x$cv_se)),
                 ...)
  graphics::segments(k_seq,
                     x$cv_error - x$cv_se,
                     k_seq,
                     x$cv_error + x$cv_se,
                     col = "grey60")
  graphics::abline(v = x$k_opt, lty = 2, col = "firebrick")
  graphics::text(x$k_opt,
                 max(x$cv_error + x$cv_se),
                 paste0("k* = ", x$k_opt),
                 adj = c(-0.1, 1), col = "firebrick", cex = 0.9)
  invisible(x)
}


# -----------------------------------------------------------------------
#  krnn_smoother
# -----------------------------------------------------------------------

#' k-RNN Moving-Average Smoother
#'
#' Estimates the conditional probability
#' \eqn{\hat{P}(y = 0 \mid \mathrm{sdp})} using the running-mean smoother
#' (moving average) with span \eqn{2k}. This is the regression interpretation
#' of the k-RNN (Section 3.4 of Perez Ruiz, 2020).
#'
#' @param X Numeric matrix (\eqn{N \times M}).
#' @param y Integer vector of labels (0/1).
#' @param k Half-neighbourhood size.
#' @param grid_eval Optional numeric vector of signed-depth values at which to
#'   evaluate the smoother. Defaults to the training signed depths.
#' @param boot Logical; if \code{TRUE} compute bootstrap-t confidence
#'   intervals. Default \code{FALSE}.
#' @param B Number of bootstrap replicates (only used when
#'   \code{boot = TRUE}). Default \code{200}.
#' @param alpha Nominal level: coverage is \eqn{1 - \alpha}. Default
#'   \code{0.05}.
#' @param grid_fd Numeric vector of length \eqn{M} (time grid).
#' @param type Depth variant.
#'
#' @return An object of class \code{"krnn_smoother"} with components:
#'   \describe{
#'     \item{sdp_eval}{Evaluation points (signed depths).}
#'     \item{prob}{Estimated conditional probabilities.}
#'     \item{ci_lower, ci_upper}{Bootstrap CI bounds (if
#'       \code{boot = TRUE}).}
#'     \item{k}{Half-neighbourhood size used.}
#'   }
#'
#' @examples
#' set.seed(5)
#' M <- 100; N <- 80
#' t  <- seq(0, 1, length.out = M)
#' X0 <- t(replicate(40, sin(2 * pi * t) + rnorm(M, sd = 0.4)))
#' X1 <- t(replicate(40, cos(2 * pi * t) + rnorm(M, sd = 0.4)))
#' X  <- rbind(X0, X1)
#' y  <- rep(0:1, each = 40)
#' sm <- krnn_smoother(X, y, k = 10)
#' plot_krnn_smoother(sm)
#'
#' @export
krnn_smoother <- function(X, y,
                          k         = 10L,
                          grid_eval = NULL,
                          boot      = FALSE,
                          B         = 200L,
                          alpha     = 0.05,
                          grid_fd   = NULL,
                          type      = c("tukey", "simplicial")) {
  type <- match.arg(type)
  y    <- as.integer(y)
  M    <- ncol(X)
  if (is.null(grid_fd)) grid_fd <- seq(0, 1, length.out = M)
  sd_obj <- signed_depth(X, grid = grid_fd, type = type)
  sdp    <- sd_obj$sdp
  o      <- order(sdp)
  sdp_s  <- sdp[o]; y_s <- y[o]
  if (is.null(grid_eval)) grid_eval <- sdp_s
  prob <- .smoother_prob(grid_eval, sdp_s, y_s, k, pre_sorted = TRUE)
  out  <- list(sdp_eval = grid_eval, prob = prob, k = k,
               sd_obj = sd_obj, y = y)
  if (boot) {
    ci           <- .bootstrap_ci(sdp_s, y_s, k, B = B, alpha = alpha)
    out$ci_lower <- ci$lower
    out$ci_upper <- ci$upper
  }
  structure(out, class = "krnn_smoother")
}


# -----------------------------------------------------------------------
#  plot_krnn_smoother
# -----------------------------------------------------------------------

#' Plot a krnn_smoother Object
#'
#' Plots the estimated conditional probability
#' \eqn{\hat{P}(y = 0 \mid \mathrm{sdp})} against the signed depth, with
#' optional bootstrap confidence bands.
#'
#' @param x A \code{"krnn_smoother"} object returned by
#'   \code{\link{krnn_smoother}}.
#' @param ... Additional graphical parameters passed to
#'   \code{\link[graphics]{plot.default}}.
#'
#' @return Invisibly returns \code{x}.
#'
#' @examples
#' set.seed(5)
#' M <- 100; N <- 80
#' t  <- seq(0, 1, length.out = M)
#' X0 <- t(replicate(40, sin(2 * pi * t) + rnorm(M, sd = 0.4)))
#' X1 <- t(replicate(40, cos(2 * pi * t) + rnorm(M, sd = 0.4)))
#' X  <- rbind(X0, X1)
#' y  <- rep(0:1, each = 40)
#' sm <- krnn_smoother(X, y, k = 10)
#' plot_krnn_smoother(sm)
#'
#' @export
plot_krnn_smoother <- function(x, ...) {
  graphics::plot(x$sdp_eval, x$prob,
                 type = "l", lwd = 2,
                 xlab = "Signed depth",
                 ylab = expression(hat(P)(y == 0 ~ "|" ~ sdp)),
                 main = paste0("k-RNN smoother  (k = ", x$k, ")"),
                 ylim = c(0, 1),
                 ...)
  graphics::abline(h = 0.5, lty = 2, col = "grey50")
  if (!is.null(x$ci_lower)) {
    graphics::lines(x$sdp_eval, x$ci_lower, lty = 3, col = "steelblue")
    graphics::lines(x$sdp_eval, x$ci_upper, lty = 3, col = "steelblue")
  }
  invisible(x)
}

test_that("fm_depth returns correct length and range (Tukey)", {
  set.seed(1)
  X <- matrix(rnorm(50 * 100), nrow = 50)
  d <- fm_depth(X, type = "tukey")
  expect_length(d, 50)
  expect_true(all(d >= 0))
  expect_true(all(d <= 1))
})

test_that("fm_depth returns correct range for simplicial variant", {
  set.seed(2)
  X <- matrix(rnorm(40 * 80), nrow = 40)
  d <- fm_depth(X, type = "simplicial")
  expect_length(d, 40)
  expect_true(all(d >= 0.5 - 1e-9))
  expect_true(all(d <= 1 + 1e-9))
})

test_that("fm_depth errors on non-matrix input", {
  expect_error(fm_depth(1:10), "'X' must be a numeric matrix")
})

test_that("fm_depth errors on mismatched grid", {
  X <- matrix(rnorm(20 * 50), nrow = 20)
  expect_error(fm_depth(X, grid = seq(0, 1, length.out = 40)),
               "'grid' must have the same length")
})

test_that("reference_curve returns a vector of correct length", {
  set.seed(3)
  X <- matrix(rnorm(30 * 60), nrow = 30)
  ref <- reference_curve(X)
  expect_length(ref, 60)
})

test_that("signed_distance_integral has correct sign structure", {
  M <- 60
  t <- seq(0, 1, length.out = M)
  X_above <- matrix(rep( 1, M), nrow = 1)
  X_below <- matrix(rep(-1, M), nrow = 1)
  X_zero  <- matrix(rep( 0, M), nrow = 1)
  X   <- rbind(X_above, X_below, X_zero)
  sdi <- signed_distance_integral(X, x_ref = X_zero[1, ], grid = t)
  expect_gt(sdi[1],  0)
  expect_lt(sdi[2],  0)
  expect_equal(sdi[3], 0, tolerance = 1e-10)
})

test_that("signed_depth returns fd_signed_depth object", {
  set.seed(7)
  X <- matrix(rnorm(50 * 80), nrow = 50)
  sd_obj <- signed_depth(X)
  expect_s3_class(sd_obj, "fd_signed_depth")
  expect_length(sd_obj$sdp, 50)
  expect_length(sd_obj$x_ref, 80)
})

test_that("krnn_fit returns predicted labels of correct length", {
  dat <- simulate_fd(n0 = 30, n1 = 30, M = 60, seed = 42)
  fit <- krnn_fit(dat$X[1:40, ], dat$y[1:40], dat$X[41:60, ], k = 5)
  expect_length(fit$predicted, 20)
  expect_true(all(fit$predicted %in% c(0L, 1L)))
})

test_that("krnn_cv selects a valid k", {
  dat <- simulate_fd(n0 = 30, n1 = 30, M = 60, seed = 1)
  cv  <- krnn_cv(dat$X, dat$y, k_max = 10, R = 3, seed = 1)
  expect_s3_class(cv, "krnn_cv")
  expect_true(cv$k_opt >= 1L && cv$k_opt <= 10L)
  expect_length(cv$cv_error, 10)
})

test_that("plot_krnn_cv runs without error", {
  dat <- simulate_fd(n0 = 30, n1 = 30, M = 60, seed = 1)
  cv  <- krnn_cv(dat$X, dat$y, k_max = 5, R = 3, seed = 1)
  expect_invisible(plot_krnn_cv(cv))
})

test_that("krnn_smoother returns krnn_smoother object", {
  dat <- simulate_fd(n0 = 30, n1 = 30, M = 60, seed = 5)
  sm  <- krnn_smoother(dat$X, dat$y, k = 5)
  expect_s3_class(sm, "krnn_smoother")
  expect_length(sm$prob, length(sm$sdp_eval))
})

test_that("plot_krnn_smoother runs without error", {
  dat <- simulate_fd(n0 = 30, n1 = 30, M = 60, seed = 5)
  sm  <- krnn_smoother(dat$X, dat$y, k = 5)
  expect_invisible(plot_krnn_smoother(sm))
})

test_that("bayes_depth_classify produces valid probabilities", {
  dat <- simulate_fd(n0 = 40, n1 = 40, M = 60, seed = 2)
  fit <- bayes_depth_classify(dat$X[1:60, ], dat$y[1:60], dat$X[61:80, ])
  expect_true(all(fit$prob >= 0 & fit$prob <= 1 + 1e-9))
  expect_true(all(abs(rowSums(fit$prob) - 1) < 1e-9))
})

test_that("logistic_depth_classify: model sdp runs without error", {
  dat <- simulate_fd(n0 = 40, n1 = 40, M = 60, seed = 3)
  fit <- logistic_depth_classify(dat$X[1:60, ], dat$y[1:60],
                                 dat$X[61:80, ], model = "sdp")
  expect_true(all(fit$predicted %in% c(0L, 1L)))
})

test_that("gam_depth_classify returns valid predictions", {
  dat <- simulate_fd(n0 = 40, n1 = 40, M = 60, seed = 4)
  fit <- gam_depth_classify(dat$X[1:60, ], dat$y[1:60], dat$X[61:80, ])
  expect_true(all(fit$predicted %in% c(0L, 1L)))
  expect_s3_class(fit, "fd_gam_fit")
})

test_that("simulate_fd dimensions are correct", {
  dat <- simulate_fd(n0 = 20, n1 = 30, M = 50, seed = 99)
  expect_equal(nrow(dat$X), 50)
  expect_equal(ncol(dat$X), 50)
  expect_equal(sum(dat$y == 0), 20)
  expect_equal(sum(dat$y == 1), 30)
})

# ============================================================
#  classifiers.R
#  bayes_depth_classify()    — Section 3.6
#  logistic_depth_classify() — Section 3.7
#  gam_depth_classify()      — Sections 3.8-3.9
# ============================================================


# -----------------------------------------------------------------------
#  bayes_depth_classify
# -----------------------------------------------------------------------

#' Bayesian Kernel-Density Classifier on Signed Depth
#'
#' Classifies functional curves via Bayes' rule applied to the signed depths.
#' Class-conditional densities \eqn{f_g(\mathrm{sdp})} are estimated by
#' kernel density estimation; prior probabilities are estimated from class
#' frequencies or supplied by the user.
#'
#' The posterior probability that a new observation belongs to class \eqn{g}
#' is
#' \deqn{P(g \mid \mathrm{sdp}_0) =
#'   \frac{f_g(\mathrm{sdp}_0)\,\pi_g}
#'   {\sum_{g'} f_{g'}(\mathrm{sdp}_0)\,\pi_{g'}}.}
#'
#' @param X_train Numeric matrix (\eqn{n_{\mathrm{train}} \times M}).
#' @param y_train Integer vector of labels (0/1).
#' @param X_test Numeric matrix of test curves. If \code{NULL} the training
#'   set is used.
#' @param priors Named numeric vector with elements \code{"0"} and \code{"1"}
#'   giving prior probabilities. Defaults to empirical class proportions.
#' @param bw_method Bandwidth selection method passed to
#'   \code{\link[stats]{density}}: \code{"nrd0"} (default), \code{"nrd"},
#'   \code{"ucv"}, \code{"bcv"}, or \code{"SJ"}.
#' @param grid Numeric vector of length \eqn{M} (time grid).
#' @param type Depth variant.
#'
#' @return An object of class \code{"fd_bayes_fit"} with components:
#'   \describe{
#'     \item{predicted}{Predicted class labels.}
#'     \item{prob}{Matrix with columns \code{prob_0} and \code{prob_1}:
#'       posterior probabilities.}
#'     \item{log_odds}{Log-odds
#'       \eqn{\log(f_0\pi_0 / f_1\pi_1)}.}
#'     \item{sd_train}{The \code{"fd_signed_depth"} object for the training
#'       set.}
#'   }
#'
#' @examples
#' set.seed(11)
#' M <- 80; N <- 100
#' t  <- seq(0, 1, length.out = M)
#' X0 <- t(replicate(50, sin(2 * pi * t) + rnorm(M, sd = 0.4)))
#' X1 <- t(replicate(50, cos(2 * pi * t) + rnorm(M, sd = 0.4)))
#' X  <- rbind(X0, X1)
#' y  <- rep(0:1, each = 50)
#' fit <- bayes_depth_classify(X, y)
#' table(fit$predicted, y)
#'
#' @export
bayes_depth_classify <- function(X_train, y_train,
                                 X_test    = NULL,
                                 priors    = NULL,
                                 bw_method = "nrd0",
                                 grid      = NULL,
                                 type      = c("tukey", "simplicial")) {
  type    <- match.arg(type)
  y_train <- as.integer(y_train)
  M       <- ncol(X_train)
  if (is.null(grid)) grid <- seq(0, 1, length.out = M)
  sd_tr  <- signed_depth(X_train, grid = grid, type = type)
  sdp_tr <- sd_tr$sdp
  sdp_te <- if (is.null(X_test)) sdp_tr else
    .project_signed_depth(X_test, sd_tr)
  classes <- sort(unique(y_train))
  if (is.null(priors)) {
    priors <- vapply(classes, function(g) mean(y_train == g), numeric(1L))
    names(priors) <- as.character(classes)
  }
  kdes <- lapply(classes, function(g)
    stats::density(sdp_tr[y_train == g], bw = bw_method))
  names(kdes) <- as.character(classes)
  eval_kde <- function(kde_obj, x)
    stats::approx(kde_obj$x, kde_obj$y, xout = x, rule = 2)$y
  dens_mat <- sapply(as.character(classes), function(g)
    eval_kde(kdes[[g]], sdp_te) * priors[g])
  if (is.vector(dens_mat)) dens_mat <- matrix(dens_mat, nrow = 1)
  post_mat           <- sweep(dens_mat, 1, rowSums(dens_mat), "/")
  colnames(post_mat) <- paste0("prob_", classes)
  predicted          <- classes[apply(post_mat, 1, which.max)]
  log_odds           <- log((dens_mat[, "0"] + 1e-300) /
                            (dens_mat[, "1"] + 1e-300))
  structure(
    list(predicted = predicted, prob = post_mat,
         log_odds  = log_odds,  sd_train = sd_tr),
    class = "fd_bayes_fit"
  )
}

#' @param x A \code{"fd_bayes_fit"} object.
#' @param ... Ignored.
#' @return Invisibly returns \code{x}.
#' @rdname bayes_depth_classify
#' @export
print.fd_bayes_fit <- function(x, ...) {
  cat("Bayesian Depth Classifier\n")
  cat("  Training N:", x$sd_train$N, "\n")
  invisible(x)
}


# -----------------------------------------------------------------------
#  logistic_depth_classify
# -----------------------------------------------------------------------

#' Logistic Regression Classifier on Signed Depth
#'
#' Fits a logistic regression model with class label as response and signed
#' depth (and optionally signed distance to the mode) as covariates
#' (Section 3.7 of Perez Ruiz, 2020).
#'
#' Three model formulae are supported:
#' \describe{
#'   \item{\code{"sdp"}}{Model 1: signed depth only.}
#'   \item{\code{"sdp+mode"}}{Model 2: signed depth plus signed distance to
#'     mode.}
#'   \item{\code{"sdp*mode"}}{Model 3: Model 2 plus interaction term.}
#' }
#'
#' @param X_train Numeric matrix (\eqn{n_{\mathrm{train}} \times M}).
#' @param y_train Integer vector of labels (0/1).
#' @param X_test Numeric matrix. If \code{NULL} the training set is used.
#' @param model Character: \code{"sdp"} (default), \code{"sdp+mode"}, or
#'   \code{"sdp*mode"}.
#' @param n_pc Number of principal components for mode estimation. Default 10.
#' @param grid Numeric vector of length \eqn{M}.
#' @param type Depth variant.
#'
#' @return An object of class \code{"fd_logistic_fit"} with components:
#'   \describe{
#'     \item{predicted}{Predicted class labels.}
#'     \item{prob}{Matrix with columns \code{prob_0} and \code{prob_1}.}
#'     \item{glm_fit}{The fitted \code{\link[stats]{glm}} object.}
#'     \item{sd_train}{The \code{"fd_signed_depth"} object for the training
#'       set.}
#'     \item{model}{Character: model formula used.}
#'   }
#'
#' @examples
#' set.seed(13)
#' M <- 80; N <- 100
#' t  <- seq(0, 1, length.out = M)
#' X0 <- t(replicate(50, sin(2 * pi * t) + rnorm(M, sd = 0.4)))
#' X1 <- t(replicate(50, cos(2 * pi * t) + rnorm(M, sd = 0.4)))
#' X  <- rbind(X0, X1)
#' y  <- rep(0:1, each = 50)
#' fit <- logistic_depth_classify(X, y, model = "sdp")
#' summary(fit$glm_fit)
#'
#' @export
logistic_depth_classify <- function(X_train, y_train,
                                    X_test = NULL,
                                    model  = c("sdp", "sdp+mode", "sdp*mode"),
                                    n_pc   = 10L,
                                    grid   = NULL,
                                    type   = c("tukey", "simplicial")) {
  model   <- match.arg(model)
  type    <- match.arg(type)
  y_train <- as.integer(y_train)
  M       <- ncol(X_train)
  if (is.null(grid)) grid <- seq(0, 1, length.out = M)
  sd_tr  <- signed_depth(X_train, grid = grid, type = type)
  sdp_tr <- sd_tr$sdp
  df_tr  <- data.frame(y = y_train, sdp = sdp_tr)
  sdm_ref <- NULL
  if (model %in% c("sdp+mode", "sdp*mode")) {
    sdm_tr      <- .signed_dist_to_mode(X_train, n_pc = n_pc)
    sdm_ref     <- attr(sdm_tr, "ref_mode")
    df_tr$sdm   <- sdm_tr
  }
  gf <- switch(model,
    "sdp"      = stats::glm(y ~ sdp,       data = df_tr,
                            family = stats::binomial()),
    "sdp+mode" = stats::glm(y ~ sdp + sdm, data = df_tr,
                            family = stats::binomial()),
    "sdp*mode" = stats::glm(y ~ sdp * sdm, data = df_tr,
                            family = stats::binomial())
  )
  if (is.null(X_test)) {
    df_te <- df_tr[, setdiff(names(df_tr), "y"), drop = FALSE]
  } else {
    sdp_te <- .project_signed_depth(X_test, sd_tr)
    df_te  <- data.frame(sdp = sdp_te)
    if (model %in% c("sdp+mode", "sdp*mode"))
      df_te$sdm <- .signed_dist_to_mode(X_test, n_pc = n_pc,
                                        ref_mode = sdm_ref)
  }
  prob_1    <- stats::predict(gf, newdata = df_te, type = "response")
  predicted <- as.integer(prob_1 > 0.5)
  structure(
    list(predicted = predicted,
         prob      = cbind(prob_0 = 1 - prob_1, prob_1 = prob_1),
         glm_fit   = gf, sd_train = sd_tr, model = model),
    class = "fd_logistic_fit"
  )
}

#' @param x A \code{"fd_logistic_fit"} object.
#' @param ... Ignored.
#' @return Invisibly returns \code{x}.
#' @rdname logistic_depth_classify
#' @export
print.fd_logistic_fit <- function(x, ...) {
  cat("Logistic Depth Classifier\n")
  cat("  Model      :", x$model, "\n")
  cat("  Training N :", x$sd_train$N, "\n")
  cat("  AIC        :", round(x$glm_fit$aic, 2), "\n")
  invisible(x)
}


# -----------------------------------------------------------------------
#  gam_depth_classify
# -----------------------------------------------------------------------

#' GAM Classifier on Signed Depth
#'
#' Fits a generalised additive model (GAM) with a smooth term on the signed
#' depth to estimate class membership probabilities. An optional iterative
#' outlier down-weighting scheme is available (Section 3.8 of Perez Ruiz,
#' 2020).
#'
#' @param X_train Numeric matrix (\eqn{N \times M}).
#' @param y_train Integer vector of labels (0/1).
#' @param X_test Numeric matrix. If \code{NULL} the training set is used.
#' @param covariates Character: \code{"sdp"} (default) or \code{"sdp+mode"}
#'   to include a second smooth on the signed distance to the mode.
#' @param k_gam Basis dimension for \code{\link[mgcv]{s}}. Default \code{10}.
#' @param n_pc Number of PCs for mode estimation. Default \code{10}.
#' @param downweight Logical; if \code{TRUE} iterate with outlier
#'   down-weighting. Default \code{FALSE}.
#' @param max_iter Maximum down-weighting iterations. Default \code{10}.
#' @param grid Numeric vector of length \eqn{M}.
#' @param type Depth variant.
#'
#' @return An object of class \code{"fd_gam_fit"} with components:
#'   \describe{
#'     \item{predicted}{Predicted class labels.}
#'     \item{prob}{Matrix with columns \code{prob_0} and \code{prob_1}.}
#'     \item{gam_fit}{The fitted \code{\link[mgcv]{gam}} object.}
#'     \item{sd_train}{The \code{"fd_signed_depth"} object for the training
#'       set.}
#'     \item{covariates}{Character: covariates used.}
#'   }
#'
#' @examples
#' set.seed(17)
#' M <- 80; N <- 100
#' t  <- seq(0, 1, length.out = M)
#' X0 <- t(replicate(50, sin(2 * pi * t) + rnorm(M, sd = 0.4)))
#' X1 <- t(replicate(50, cos(2 * pi * t) + rnorm(M, sd = 0.4)))
#' X  <- rbind(X0, X1)
#' y  <- rep(0:1, each = 50)
#' fit <- gam_depth_classify(X, y)
#' print(fit)
#'
#' @export
gam_depth_classify <- function(X_train, y_train,
                               X_test     = NULL,
                               covariates = c("sdp", "sdp+mode"),
                               k_gam      = 10L,
                               n_pc       = 10L,
                               downweight = FALSE,
                               max_iter   = 10L,
                               grid       = NULL,
                               type       = c("tukey", "simplicial")) {
  covariates <- match.arg(covariates)
  type       <- match.arg(type)
  y_train    <- as.integer(y_train)
  classes    <- sort(unique(y_train))
  G          <- length(classes)
  M          <- ncol(X_train)
  if (is.null(grid)) grid <- seq(0, 1, length.out = M)
  if (G != 2L)
    stop("Multiclass GAM is planned for v0.2. ",
         "For G > 2 use logistic_depth_classify() with pairwise comparisons.")
  sd_tr  <- signed_depth(X_train, grid = grid, type = type)
  sdp_tr <- sd_tr$sdp
  df_tr  <- data.frame(y_bin = as.integer(y_train == classes[2L]),
                       sdp   = sdp_tr)
  sdm_ref <- NULL
  if (covariates == "sdp+mode") {
    sdm_tr      <- .signed_dist_to_mode(X_train, n_pc = n_pc)
    sdm_ref     <- attr(sdm_tr, "ref_mode")
    df_tr$sdm   <- sdm_tr
  }
  if (covariates == "sdp") {
    gam_formula <- y_bin ~ s(sdp, k = 10)
  } else {
    gam_formula <- y_bin ~ s(sdp, k = 10) + s(sdm, k = 10)
  }
  weights_vec <- rep(1, nrow(df_tr))
  if (downweight) {
    for (iter in seq_len(max_iter)) {
      environment(gam_formula) <- environment()
      gam_fit  <- mgcv::gam(gam_formula, data = df_tr,
                            family = stats::binomial(),
                            weights = weights_vec)
      fitted_p <- stats::fitted(gam_fit)
      resid_p  <- abs(df_tr$y_bin - fitted_p)
      weights_vec <- ifelse(resid_p > 0.5, 0.5 / resid_p, 1)
    }
  } else {
    environment(gam_formula) <- environment()
    gam_fit <- mgcv::gam(gam_formula, data = df_tr,
                         family = stats::binomial())
  }
  if (is.null(X_test)) {
    df_te <- df_tr[, setdiff(names(df_tr), "y_bin"), drop = FALSE]
  } else {
    sdp_te <- .project_signed_depth(X_test, sd_tr)
    df_te  <- data.frame(sdp = sdp_te)
    if (covariates == "sdp+mode")
      df_te$sdm <- .signed_dist_to_mode(X_test, n_pc = n_pc,
                                        ref_mode = sdm_ref)
  }
  prob_1    <- stats::predict(gam_fit, newdata = df_te, type = "response")
  predicted <- as.integer(prob_1 > 0.5)
  structure(
    list(predicted  = predicted,
         prob       = cbind(prob_0 = 1 - prob_1, prob_1 = prob_1),
         gam_fit    = gam_fit, sd_train = sd_tr, covariates = covariates),
    class = "fd_gam_fit"
  )
}

#' @param x A \code{"fd_gam_fit"} object.
#' @param ... Ignored.
#' @return Invisibly returns \code{x}.
#' @rdname gam_depth_classify
#' @export
print.fd_gam_fit <- function(x, ...) {
  cat("GAM Depth Classifier\n")
  cat("  Covariates :", x$covariates, "\n")
  cat("  Training N :", x$sd_train$N, "\n")
  cat("  Dev. expl. :", round(summary(x$gam_fit)$dev.expl * 100, 1), "%\n")
  invisible(x)
}


# -----------------------------------------------------------------------
#  Internal: signed distance to the modal curve (Section 3.7.1)
# -----------------------------------------------------------------------

#' @keywords internal
.signed_dist_to_mode <- function(X, n_pc = 10L, ref_mode = NULL) {
  N   <- nrow(X); M <- ncol(X)
  npc <- min(as.integer(n_pc), N - 1L, M)
  X_c     <- scale(X, center = TRUE, scale = FALSE)
  svd_res <- svd(X_c / sqrt(N - 1), nu = 0, nv = npc)
  scores  <- X_c %*% svd_res$v
  if (is.null(ref_mode)) {
    modes <- vapply(seq_len(npc), function(j)
      modeest::mfv1(scores[, j], method = "parzen"), numeric(1L))
  } else {
    modes <- ref_mode
  }
  sdm <- as.numeric(
    (scores - matrix(modes, nrow = N, ncol = npc, byrow = TRUE)) %*%
    rep(1, npc))
  attr(sdm, "ref_mode") <- modes
  sdm
}

#' Simulate a Two-Group Functional Dataset
#'
#' Generates a labelled functional dataset from two Gaussian processes,
#' useful for illustrating and testing the classifiers in `fdclassify`.
#' The two groups differ in their mean function:
#' \deqn{
#'   x_i(t) = \mu_g(t) + \varepsilon_i(t), \quad
#'   \varepsilon_i(t) \sim \mathcal{GP}(0, \sigma^2),
#' }
#' where \eqn{\mu_0(t) = A_0 \sin(2\pi f_0 t)} and
#' \eqn{\mu_1(t) = A_1 \cos(2\pi f_1 t)} by default.
#'
#' @param n0 Number of curves in group 0.  Default 50.
#' @param n1 Number of curves in group 1.  Default 50.
#' @param M  Number of time points.  Default 100.
#' @param sigma Standard deviation of the noise.  Default 0.4.
#' @param A0,A1 Amplitudes of the mean functions.  Both default to 1.
#' @param f0,f1 Frequencies of the mean functions.  Both default to 1.
#' @param seed Optional integer seed.
#'
#' @return A list with:
#'   \describe{
#'     \item{`X`}{Numeric matrix \eqn{(n_0+n_1) \times M}.}
#'     \item{`y`}{Integer vector of labels (0 or 1).}
#'     \item{`grid`}{Numeric vector of length \eqn{M} on \eqn{[0,1]}.}
#'   }
#'
#' @examples
#' dat <- simulate_fd(n0 = 40, n1 = 40, seed = 1)
#' matplot(t(dat$X[dat$y == 0, ]), type = "l", col = "steelblue",
#'         lty = 1, xlab = "t", ylab = "x(t)", main = "Simulated curves")
#' matlines(t(dat$X[dat$y == 1, ]), col = "firebrick", lty = 1)
#' legend("topright", legend = c("Group 0", "Group 1"),
#'        col = c("steelblue","firebrick"), lty = 1)
#'
#' @export
simulate_fd <- function(n0    = 50L,
                        n1    = 50L,
                        M     = 100L,
                        sigma = 0.4,
                        A0    = 1,
                        A1    = 1,
                        f0    = 1,
                        f1    = 1,
                        seed  = NULL) {
  if (!is.null(seed)) set.seed(seed)
  grid <- seq(0, 1, length.out = M)
  mu0  <- A0 * sin(2 * pi * f0 * grid)
  mu1  <- A1 * cos(2 * pi * f1 * grid)
  X0   <- t(replicate(n0, mu0 + stats::rnorm(M, sd = sigma)))
  X1   <- t(replicate(n1, mu1 + stats::rnorm(M, sd = sigma)))
  list(
    X    = rbind(X0, X1),
    y    = rep(0:1, c(n0, n1)),
    grid = grid
  )
}

# ============================================================
#  signed_depth.R
#  1. reference_curve()          — median curve from combined sample
#  2. signed_distance_integral() — area between each curve and reference
#  3. signed_depth()             — main transformation for all classifiers
# ============================================================


#' Reference (Median) Curve
#'
#' Returns the curve (or average of curves) that attains the maximum
#' Fraiman-Muniz depth over the combined sample — the functional analogue
#' of the median.
#'
#' @param X Numeric matrix (\eqn{N \times M}) of curves.
#' @param depth Numeric vector of length \eqn{N} of pre-computed FM depth
#'   values. If \code{NULL} (default), \code{\link{fm_depth}} is called
#'   internally with \code{type = "tukey"}.
#' @param ... Additional arguments forwarded to \code{\link{fm_depth}} when
#'   \code{depth} is \code{NULL}.
#'
#' @return A numeric vector of length \eqn{M}.
#'
#' @examples
#' set.seed(1)
#' X <- matrix(rnorm(50 * 100), nrow = 50)
#' ref <- reference_curve(X)
#' plot(ref, type = "l", xlab = "t", ylab = "x(t)",
#'      main = "Reference (median) curve")
#'
#' @export
reference_curve <- function(X, depth = NULL, ...) {
  if (is.null(depth)) depth <- fm_depth(X, ...)
  max_depth <- max(depth)
  idx       <- which(depth == max_depth)
  if (length(idx) == 1L) return(X[idx, , drop = TRUE])
  colMeans(X[idx, , drop = FALSE])
}


#' Signed Distance Integral
#'
#' For each curve \eqn{x_i(t)}, computes
#' \deqn{\int_I \bigl(x_i(t) - x_{\mathrm{ref}}(t)\bigr)\,dt,}
#' which is positive when the curve lies predominantly above the reference
#' and negative when it lies below. This integral assigns a sign to the depth
#' of each curve (equation 3.1 of Perez Ruiz, 2020).
#'
#' @param X Numeric matrix (\eqn{N \times M}).
#' @param x_ref Numeric vector of length \eqn{M}: the reference curve.
#'   If \code{NULL} (default), \code{\link{reference_curve}} is called
#'   internally.
#' @param grid Numeric vector of length \eqn{M} with observation times.
#'   Defaults to equally spaced points on \eqn{[0, 1]}.
#'
#' @return A numeric vector of length \eqn{N}.
#'
#' @seealso \code{\link{signed_depth}}, \code{\link{reference_curve}}
#'
#' @examples
#' set.seed(1)
#' X <- matrix(rnorm(40 * 80), nrow = 40)
#' sdi <- signed_distance_integral(X)
#' hist(sdi, main = "Signed distance integrals", xlab = "SDI")
#'
#' @export
signed_distance_integral <- function(X, x_ref = NULL, grid = NULL) {
  if (!is.matrix(X) || !is.numeric(X))
    stop("'X' must be a numeric matrix.")
  M <- ncol(X)
  if (is.null(grid))  grid  <- seq(0, 1, length.out = M)
  if (is.null(x_ref)) x_ref <- reference_curve(X, grid = grid)
  if (length(x_ref) != M)
    stop("'x_ref' must have the same length as ncol(X).")
  widths   <- diff(grid)
  diff_mat <- sweep(X, 2, x_ref, "-")
  as.numeric(diff_mat[, -1, drop = FALSE] %*% widths)
}


#' Signed Depth
#'
#' The core transformation of the k-RNN pipeline. For each curve computes
#' \deqn{\mathrm{sdp}(x_i(t)) =
#'   \mathrm{sgn}\!\left\{\int_I(x_i(t)-x_{\mathrm{ref}}(t))\,dt\right\}
#'   \times D(x_i(t)),}
#' where \eqn{D(\cdot)} is the Fraiman-Muniz depth and the sign is derived
#' from the signed distance integral (equation 3.2 of Perez Ruiz, 2020).
#' Curves above the reference receive a positive signed depth; curves below
#' receive a negative one.
#'
#' @param X Numeric matrix (\eqn{N \times M}) with curves in rows.
#' @param grid Numeric vector of length \eqn{M}. Defaults to equally spaced
#'   points on \eqn{[0, 1]}.
#' @param type Depth variant: \code{"tukey"} (default) or
#'   \code{"simplicial"}.
#' @param x_ref Optional pre-computed reference curve (numeric vector of
#'   length \eqn{M}). Computed internally if \code{NULL}.
#' @param depth Optional pre-computed FM depth vector of length \eqn{N}.
#'   Computed internally if \code{NULL}.
#'
#' @return An object of class \code{"fd_signed_depth"} with components:
#'   \describe{
#'     \item{sdp}{Numeric vector of length \eqn{N}: the signed depths.}
#'     \item{depth}{Numeric vector of length \eqn{N}: raw FM depths.}
#'     \item{sdi}{Numeric vector of length \eqn{N}: signed distance
#'       integrals.}
#'     \item{x_ref}{Numeric vector of length \eqn{M}: reference curve.}
#'     \item{grid}{Numeric vector of length \eqn{M}: time grid used.}
#'     \item{type}{Character: depth variant used.}
#'     \item{N}{Number of curves.}
#'     \item{M}{Number of time points.}
#'   }
#'
#' @references
#' Perez Ruiz, D. A. (2020). \emph{Supervised Classification for Functional
#' Data}. PhD thesis, University of Manchester.
#'
#' @examples
#' set.seed(42)
#' N <- 60; M <- 100
#' t  <- seq(0, 1, length.out = M)
#' X0 <- t(replicate(N / 2, sin(2 * pi * t) + rnorm(M, sd = 0.3)))
#' X1 <- t(replicate(N / 2, cos(2 * pi * t) + rnorm(M, sd = 0.3)))
#' X  <- rbind(X0, X1)
#' sd_obj <- signed_depth(X)
#' plot(sd_obj$sdp, col = rep(1:2, each = N / 2),
#'      xlab = "Curve index", ylab = "Signed depth",
#'      main = "Signed depth by group")
#' abline(h = 0, lty = 2)
#'
#' @export
signed_depth <- function(X, grid = NULL, type = c("tukey", "simplicial"),
                         x_ref = NULL, depth = NULL) {
  type <- match.arg(type)
  if (!is.matrix(X) || !is.numeric(X))
    stop("'X' must be a numeric matrix with curves in rows.")
  N <- nrow(X); M <- ncol(X)
  if (is.null(grid))  grid  <- seq(0, 1, length.out = M)
  if (is.null(depth)) depth <- fm_depth(X, grid = grid, type = type)
  if (is.null(x_ref)) x_ref <- reference_curve(X, depth = depth)
  sdi <- signed_distance_integral(X, x_ref = x_ref, grid = grid)
  sdp <- sign(sdi) * depth
  structure(
    list(sdp = sdp, depth = depth, sdi = sdi,
         x_ref = x_ref, grid = grid, type = type, N = N, M = M),
    class = "fd_signed_depth"
  )
}

#' @param x A \code{"fd_signed_depth"} object.
#' @param ... Ignored.
#' @return Invisibly returns \code{x}.
#' @rdname signed_depth
#' @export
print.fd_signed_depth <- function(x, ...) {
  cat("Signed depth summary\n")
  cat("  Curves (N)     :", x$N, "\n")
  cat("  Time points (M):", x$M, "\n")
  cat("  Depth type     :", x$type, "\n")
  cat("  sdp range      : [", round(min(x$sdp), 4), ",",
      round(max(x$sdp), 4), "]\n")
  invisible(x)
}

#' Fraiman-Muniz Functional Depth
#'
#' Computes the Fraiman-Muniz (FM) depth for a collection of discretised
#' functional observations. Two variants are supported: the Tukey-FM depth,
#' which maps depth values to \eqn{[0, 1]}, and the Simplicial-FM depth,
#' which maps to \eqn{[0.5, 1]}.  The two are related by
#' \deqn{\text{Tukey-FM}_i = 2\,(\text{Simplicial-FM}_i - 1/2).}
#'
#' For a discretised dataset \eqn{\{x_i(t_j)\}}, \eqn{i = 1,\ldots,N},
#' \eqn{j = 1,\ldots,M}, the Simplicial-FM depth is
#' \deqn{
#'   \text{FM}_i = \sum_{j=2}^{M}(t_j - t_{j-1})
#'   \left[1 - \left|\frac{1}{2} - F_{N,t_j}(x_i(t_j))\right|\right],
#' }
#' where \eqn{F_{N,t}} is the empirical CDF of the sample at time \eqn{t}.
#' The Tukey-FM depth substitutes the Tukey univariate depth
#' \eqn{\min\{F_{N,t}(x), 1 - F_{N,t}(x)\}} at each time point.
#'
#' @param X Numeric matrix of dimension \eqn{N \times M}, where rows are curves
#'   and columns are time points.  Missing values are not allowed.
#' @param grid Numeric vector of length \eqn{M} giving the observation times
#'   (e.g. `seq(0, 1, length.out = M)`).  If `NULL` (default) equally spaced
#'   points on \eqn{[0, 1]} are assumed.
#' @param type Character string, either `"tukey"` (default) or `"simplicial"`,
#'   selecting the univariate depth used at each time point.
#'
#' @return A numeric vector of length \eqn{N} containing the FM depth value
#'   for each curve.
#'
#' @references
#' Fraiman, R. and Muniz, G. (2001). Trimmed means for functional data.
#' *Test*, **10**(2), 419–440.
#'
#' López-Pintado, S. and Romo, J. (2009). On the concept of depth for
#' functional data. *Journal of the American Statistical Association*,
#' **104**(486), 718–734.
#'
#' @examples
#' set.seed(1)
#' N <- 50; M <- 100
#' X <- matrix(rnorm(N * M), nrow = N)
#' d <- fm_depth(X)
#' plot(d, xlab = "Curve index", ylab = "Tukey-FM depth")
#'
#' @export
fm_depth <- function(X, grid = NULL, type = c("tukey", "simplicial")) {
  type <- match.arg(type)

  if (!is.matrix(X) || !is.numeric(X))
    stop("'X' must be a numeric matrix with curves in rows and time points in columns.")
  if (anyNA(X))
    stop("'X' contains missing values.  Impute or remove them before calling fm_depth().")

  N <- nrow(X)
  M <- ncol(X)

  if (is.null(grid)) {
    grid <- seq(0, 1, length.out = M)
  } else {
    if (length(grid) != M)
      stop("'grid' must have the same length as ncol(X).")
    if (is.unsorted(grid))
      stop("'grid' must be non-decreasing.")
  }

  # Widths for the trapezoidal / left-endpoint sum  (t_j - t_{j-1})
  widths <- diff(grid)                   # length M-1

  # At each time point t_j, compute the empirical CDF F_{N,t}(x_i(t_j))
  # for all i simultaneously.
  # ecdf_mat[i, j] = (1/N) * #{k : X[k, j] <= X[i, j]}
  ecdf_mat <- matrix(0, nrow = N, ncol = M)
  for (j in seq_len(M)) {
    col_j <- X[, j]
    # rank() with ties.method="max" gives #{k : x_k <= x_i}
    ecdf_mat[, j] <- rank(col_j, ties.method = "max") / N
  }

  # Univariate depth at each time point
  if (type == "tukey") {
    depth_t <- pmin(ecdf_mat, 1 - ecdf_mat)          # Tukey: min(F, 1-F)
  } else {
    depth_t <- 1 - abs(0.5 - ecdf_mat)               # Simplicial: 1 - |0.5 - F|
  }

  # Integrate using the left-endpoint rule over the grid
  # depth_t has columns j = 1..M; we sum columns 2..M weighted by widths
  fm <- as.numeric(depth_t[, -1, drop = FALSE] %*% widths)

  fm
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/signed_depth.R
\name{reference_curve}
\alias{reference_curve}
\title{Reference (Median) Curve}
\usage{
reference_curve(X, depth = NULL, ...)
}
\arguments{
\item{X}{Numeric matrix (\eqn{N \times M}) of curves.}

\item{depth}{Numeric vector of length \eqn{N} of pre-computed FM depth
values. If \code{NULL} (default), \code{\link{fm_depth}} is called
internally with \code{type = "tukey"}.}

\item{...}{Additional arguments forwarded to \code{\link{fm_depth}} when
\code{depth} is \code{NULL}.}
}
\value{
A numeric vector of length \eqn{M}.
}
\description{
Returns the curve (or average of curves) that attains the maximum
Fraiman-Muniz depth over the combined sample — the functional analogue
of the median.
}
\examples{
set.seed(1)
X <- matrix(rnorm(50 * 100), nrow = 50)
ref <- reference_curve(X)
plot(ref, type = "l", xlab = "t", ylab = "x(t)",
     main = "Reference (median) curve")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/classifiers.R
\name{gam_depth_classify}
\alias{gam_depth_classify}
\alias{print.fd_gam_fit}
\title{GAM Classifier on Signed Depth}
\usage{
gam_depth_classify(
  X_train,
  y_train,
  X_test = NULL,
  covariates = c("sdp", "sdp+mode"),
  k_gam = 10L,
  n_pc = 10L,
  downweight = FALSE,
  max_iter = 10L,
  grid = NULL,
  type = c("tukey", "simplicial")
)

\method{print}{fd_gam_fit}(x, ...)
}
\arguments{
\item{X_train}{Numeric matrix (\eqn{N \times M}).}

\item{y_train}{Integer vector of labels (0/1).}

\item{X_test}{Numeric matrix. If \code{NULL} the training set is used.}

\item{covariates}{Character: \code{"sdp"} (default) or \code{"sdp+mode"}
to include a second smooth on the signed distance to the mode.}

\item{k_gam}{Basis dimension for \code{\link[mgcv]{s}}. Default \code{10}.}

\item{n_pc}{Number of PCs for mode estimation. Default \code{10}.}

\item{downweight}{Logical; if \code{TRUE} iterate with outlier
down-weighting. Default \code{FALSE}.}

\item{max_iter}{Maximum down-weighting iterations. Default \code{10}.}

\item{grid}{Numeric vector of length \eqn{M}.}

\item{type}{Depth variant.}

\item{x}{A \code{"fd_gam_fit"} object.}

\item{...}{Ignored.}
}
\value{
An object of class \code{"fd_gam_fit"} with components:
\describe{
\item{predicted}{Predicted class labels.}
\item{prob}{Matrix with columns \code{prob_0} and \code{prob_1}.}
\item{gam_fit}{The fitted \code{\link[mgcv]{gam}} object.}
\item{sd_train}{The \code{"fd_signed_depth"} object for the training
set.}
\item{covariates}{Character: covariates used.}
}

Invisibly returns \code{x}.
}
\description{
Fits a generalised additive model (GAM) with a smooth term on the signed
depth to estimate class membership probabilities. An optional iterative
outlier down-weighting scheme is available (Section 3.8 of Perez Ruiz,
2020).
}
\examples{
set.seed(17)
M <- 80; N <- 100
t  <- seq(0, 1, length.out = M)
X0 <- t(replicate(50, sin(2 * pi * t) + rnorm(M, sd = 0.4)))
X1 <- t(replicate(50, cos(2 * pi * t) + rnorm(M, sd = 0.4)))
X  <- rbind(X0, X1)
y  <- rep(0:1, each = 50)
fit <- gam_depth_classify(X, y)
print(fit)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/classifiers.R
\name{logistic_depth_classify}
\alias{logistic_depth_classify}
\alias{print.fd_logistic_fit}
\title{Logistic Regression Classifier on Signed Depth}
\usage{
logistic_depth_classify(
  X_train,
  y_train,
  X_test = NULL,
  model = c("sdp", "sdp+mode", "sdp*mode"),
  n_pc = 10L,
  grid = NULL,
  type = c("tukey", "simplicial")
)

\method{print}{fd_logistic_fit}(x, ...)
}
\arguments{
\item{X_train}{Numeric matrix (\eqn{n_{\mathrm{train}} \times M}).}

\item{y_train}{Integer vector of labels (0/1).}

\item{X_test}{Numeric matrix. If \code{NULL} the training set is used.}

\item{model}{Character: \code{"sdp"} (default), \code{"sdp+mode"}, or
\code{"sdp*mode"}.}

\item{n_pc}{Number of principal components for mode estimation. Default 10.}

\item{grid}{Numeric vector of length \eqn{M}.}

\item{type}{Depth variant.}

\item{x}{A \code{"fd_logistic_fit"} object.}

\item{...}{Ignored.}
}
\value{
An object of class \code{"fd_logistic_fit"} with components:
\describe{
\item{predicted}{Predicted class labels.}
\item{prob}{Matrix with columns \code{prob_0} and \code{prob_1}.}
\item{glm_fit}{The fitted \code{\link[stats]{glm}} object.}
\item{sd_train}{The \code{"fd_signed_depth"} object for the training
set.}
\item{model}{Character: model formula used.}
}

Invisibly returns \code{x}.
}
\description{
Fits a logistic regression model with class label as response and signed
depth (and optionally signed distance to the mode) as covariates
(Section 3.7 of Perez Ruiz, 2020).
}
\details{
Three model formulae are supported:
\describe{
\item{\code{"sdp"}}{Model 1: signed depth only.}
\item{\code{"sdp+mode"}}{Model 2: signed depth plus signed distance to
mode.}
\item{\code{"sdp*mode"}}{Model 3: Model 2 plus interaction term.}
}
}
\examples{
set.seed(13)
M <- 80; N <- 100
t  <- seq(0, 1, length.out = M)
X0 <- t(replicate(50, sin(2 * pi * t) + rnorm(M, sd = 0.4)))
X1 <- t(replicate(50, cos(2 * pi * t) + rnorm(M, sd = 0.4)))
X  <- rbind(X0, X1)
y  <- rep(0:1, each = 50)
fit <- logistic_depth_classify(X, y, model = "sdp")
summary(fit$glm_fit)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/simulate_fd.R
\name{simulate_fd}
\alias{simulate_fd}
\title{Simulate a Two-Group Functional Dataset}
\usage{
simulate_fd(
  n0 = 50L,
  n1 = 50L,
  M = 100L,
  sigma = 0.4,
  A0 = 1,
  A1 = 1,
  f0 = 1,
  f1 = 1,
  seed = NULL
)
}
\arguments{
\item{n0}{Number of curves in group 0.  Default 50.}

\item{n1}{Number of curves in group 1.  Default 50.}

\item{M}{Number of time points.  Default 100.}

\item{sigma}{Standard deviation of the noise.  Default 0.4.}

\item{A0, A1}{Amplitudes of the mean functions.  Both default to 1.}

\item{f0, f1}{Frequencies of the mean functions.  Both default to 1.}

\item{seed}{Optional integer seed.}
}
\value{
A list with:
\describe{
\item{\code{X}}{Numeric matrix \eqn{(n_0+n_1) \times M}.}
\item{\code{y}}{Integer vector of labels (0 or 1).}
\item{\code{grid}}{Numeric vector of length \eqn{M} on \eqn{[0,1]}.}
}
}
\description{
Generates a labelled functional dataset from two Gaussian processes,
useful for illustrating and testing the classifiers in \code{fdclassify}.
The two groups differ in their mean function:
\deqn{
  x_i(t) = \mu_g(t) + \varepsilon_i(t), \quad
  \varepsilon_i(t) \sim \mathcal{GP}(0, \sigma^2),
}
where \eqn{\mu_0(t) = A_0 \sin(2\pi f_0 t)} and
\eqn{\mu_1(t) = A_1 \cos(2\pi f_1 t)} by default.
}
\examples{
dat <- simulate_fd(n0 = 40, n1 = 40, seed = 1)
matplot(t(dat$X[dat$y == 0, ]), type = "l", col = "steelblue",
        lty = 1, xlab = "t", ylab = "x(t)", main = "Simulated curves")
matlines(t(dat$X[dat$y == 1, ]), col = "firebrick", lty = 1)
legend("topright", legend = c("Group 0", "Group 1"),
       col = c("steelblue","firebrick"), lty = 1)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/signed_depth.R
\name{signed_distance_integral}
\alias{signed_distance_integral}
\title{Signed Distance Integral}
\usage{
signed_distance_integral(X, x_ref = NULL, grid = NULL)
}
\arguments{
\item{X}{Numeric matrix (\eqn{N \times M}).}

\item{x_ref}{Numeric vector of length \eqn{M}: the reference curve.
If \code{NULL} (default), \code{\link{reference_curve}} is called
internally.}

\item{grid}{Numeric vector of length \eqn{M} with observation times.
Defaults to equally spaced points on \eqn{[0, 1]}.}
}
\value{
A numeric vector of length \eqn{N}.
}
\description{
For each curve \eqn{x_i(t)}, computes
\deqn{\int_I \bigl(x_i(t) - x_{\mathrm{ref}}(t)\bigr)\,dt,}
which is positive when the curve lies predominantly above the reference
and negative when it lies below. This integral assigns a sign to the depth
of each curve (equation 3.1 of Perez Ruiz, 2020).
}
\examples{
set.seed(1)
X <- matrix(rnorm(40 * 80), nrow = 40)
sdi <- signed_distance_integral(X)
hist(sdi, main = "Signed distance integrals", xlab = "SDI")

}
\seealso{
\code{\link{signed_depth}}, \code{\link{reference_curve}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/krnn.R
\name{plot_krnn_cv}
\alias{plot_krnn_cv}
\title{Plot a krnn_cv Object}
\usage{
plot_krnn_cv(x, ...)
}
\arguments{
\item{x}{A \code{"krnn_cv"} object returned by \code{\link{krnn_cv}}.}

\item{...}{Additional graphical parameters passed to
\code{\link[graphics]{plot.default}}.}
}
\value{
Invisibly returns \code{x}.
}
\description{
Plots the cross-validation error curve against the neighbourhood size k,
with standard-error bars and a vertical line at the selected optimum.
}
\examples{
\donttest{
set.seed(3)
M <- 80
t  <- seq(0, 1, length.out = M)
X0 <- t(replicate(50, sin(2 * pi * t) + rnorm(M, sd = 0.4)))
X1 <- t(replicate(50, cos(2 * pi * t) + rnorm(M, sd = 0.4)))
X  <- rbind(X0, X1)
y  <- rep(0:1, each = 50)
cv <- krnn_cv(X, y, k_max = 20, R = 5, seed = 1)
plot_krnn_cv(cv)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/krnn.R
\name{krnn_cv}
\alias{krnn_cv}
\alias{print.krnn_cv}
\title{Cross-Validation for k-RNN Neighbourhood Size}
\usage{
krnn_cv(
  X,
  y,
  k_max = NULL,
  R = 10L,
  rule = c("min", "1se"),
  grid = NULL,
  type = c("tukey", "simplicial"),
  seed = NULL
)

\method{print}{krnn_cv}(x, ...)
}
\arguments{
\item{X}{Numeric matrix (\eqn{N \times M}).}

\item{y}{Integer vector of class labels (0/1).}

\item{k_max}{Maximum value of k to evaluate. Default
\code{floor(nrow(X) / 4)}.}

\item{R}{Number of CV folds. Default \code{10}.}

\item{rule}{Character: \code{"min"} (default) or \code{"1se"}.}

\item{grid}{Numeric vector of length \eqn{M}.}

\item{type}{Depth variant.}

\item{seed}{Optional integer seed for reproducibility.}

\item{x}{A \code{"krnn_cv"} object.}

\item{...}{Ignored.}
}
\value{
An object of class \code{"krnn_cv"} with components:
\describe{
\item{k_opt}{Selected optimal k.}
\item{cv_error}{Mean CV misclassification error for each k.}
\item{cv_se}{Standard error of CV error per k.}
\item{k_max}{Maximum k evaluated.}
\item{R}{Number of folds used.}
\item{rule}{The selection rule used.}
}

Invisibly returns \code{x}.
}
\description{
Selects the optimal half-neighbourhood size \eqn{k} by \eqn{R}-fold
cross-validation, using either the minimum-error rule or the
one-standard-error (1-SE) rule (Section 3.3.1 of Perez Ruiz, 2020).
}
\examples{
\donttest{
set.seed(3)
M <- 80
t  <- seq(0, 1, length.out = M)
X0 <- t(replicate(50, sin(2 * pi * t) + rnorm(M, sd = 0.4)))
X1 <- t(replicate(50, cos(2 * pi * t) + rnorm(M, sd = 0.4)))
X  <- rbind(X0, X1)
y  <- rep(0:1, each = 50)
cv <- krnn_cv(X, y, k_max = 20, R = 5, seed = 1)
print(cv)
plot_krnn_cv(cv)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/signed_depth.R
\name{signed_depth}
\alias{signed_depth}
\alias{print.fd_signed_depth}
\title{Signed Depth}
\usage{
signed_depth(
  X,
  grid = NULL,
  type = c("tukey", "simplicial"),
  x_ref = NULL,
  depth = NULL
)

\method{print}{fd_signed_depth}(x, ...)
}
\arguments{
\item{X}{Numeric matrix (\eqn{N \times M}) with curves in rows.}

\item{grid}{Numeric vector of length \eqn{M}. Defaults to equally spaced
points on \eqn{[0, 1]}.}

\item{type}{Depth variant: \code{"tukey"} (default) or
\code{"simplicial"}.}

\item{x_ref}{Optional pre-computed reference curve (numeric vector of
length \eqn{M}). Computed internally if \code{NULL}.}

\item{depth}{Optional pre-computed FM depth vector of length \eqn{N}.
Computed internally if \code{NULL}.}

\item{x}{A \code{"fd_signed_depth"} object.}

\item{...}{Ignored.}
}
\value{
An object of class \code{"fd_signed_depth"} with components:
\describe{
\item{sdp}{Numeric vector of length \eqn{N}: the signed depths.}
\item{depth}{Numeric vector of length \eqn{N}: raw FM depths.}
\item{sdi}{Numeric vector of length \eqn{N}: signed distance
integrals.}
\item{x_ref}{Numeric vector of length \eqn{M}: reference curve.}
\item{grid}{Numeric vector of length \eqn{M}: time grid used.}
\item{type}{Character: depth variant used.}
\item{N}{Number of curves.}
\item{M}{Number of time points.}
}

Invisibly returns \code{x}.
}
\description{
The core transformation of the k-RNN pipeline. For each curve computes
\deqn{\mathrm{sdp}(x_i(t)) =
  \mathrm{sgn}\!\left\{\int_I(x_i(t)-x_{\mathrm{ref}}(t))\,dt\right\}
  \times D(x_i(t)),}
where \eqn{D(\cdot)} is the Fraiman-Muniz depth and the sign is derived
from the signed distance integral (equation 3.2 of Perez Ruiz, 2020).
Curves above the reference receive a positive signed depth; curves below
receive a negative one.
}
\examples{
set.seed(42)
N <- 60; M <- 100
t  <- seq(0, 1, length.out = M)
X0 <- t(replicate(N / 2, sin(2 * pi * t) + rnorm(M, sd = 0.3)))
X1 <- t(replicate(N / 2, cos(2 * pi * t) + rnorm(M, sd = 0.3)))
X  <- rbind(X0, X1)
sd_obj <- signed_depth(X)
plot(sd_obj$sdp, col = rep(1:2, each = N / 2),
     xlab = "Curve index", ylab = "Signed depth",
     main = "Signed depth by group")
abline(h = 0, lty = 2)

}
\references{
Perez Ruiz, D. A. (2020). \emph{Supervised Classification for Functional
Data}. PhD thesis, University of Manchester.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/krnn.R
\name{krnn_smoother}
\alias{krnn_smoother}
\title{k-RNN Moving-Average Smoother}
\usage{
krnn_smoother(
  X,
  y,
  k = 10L,
  grid_eval = NULL,
  boot = FALSE,
  B = 200L,
  alpha = 0.05,
  grid_fd = NULL,
  type = c("tukey", "simplicial")
)
}
\arguments{
\item{X}{Numeric matrix (\eqn{N \times M}).}

\item{y}{Integer vector of labels (0/1).}

\item{k}{Half-neighbourhood size.}

\item{grid_eval}{Optional numeric vector of signed-depth values at which to
evaluate the smoother. Defaults to the training signed depths.}

\item{boot}{Logical; if \code{TRUE} compute bootstrap-t confidence
intervals. Default \code{FALSE}.}

\item{B}{Number of bootstrap replicates (only used when
\code{boot = TRUE}). Default \code{200}.}

\item{alpha}{Nominal level: coverage is \eqn{1 - \alpha}. Default
\code{0.05}.}

\item{grid_fd}{Numeric vector of length \eqn{M} (time grid).}

\item{type}{Depth variant.}
}
\value{
An object of class \code{"krnn_smoother"} with components:
\describe{
\item{sdp_eval}{Evaluation points (signed depths).}
\item{prob}{Estimated conditional probabilities.}
\item{ci_lower, ci_upper}{Bootstrap CI bounds (if
\code{boot = TRUE}).}
\item{k}{Half-neighbourhood size used.}
}
}
\description{
Estimates the conditional probability
\eqn{\hat{P}(y = 0 \mid \mathrm{sdp})} using the running-mean smoother
(moving average) with span \eqn{2k}. This is the regression interpretation
of the k-RNN (Section 3.4 of Perez Ruiz, 2020).
}
\examples{
set.seed(5)
M <- 100; N <- 80
t  <- seq(0, 1, length.out = M)
X0 <- t(replicate(40, sin(2 * pi * t) + rnorm(M, sd = 0.4)))
X1 <- t(replicate(40, cos(2 * pi * t) + rnorm(M, sd = 0.4)))
X  <- rbind(X0, X1)
y  <- rep(0:1, each = 40)
sm <- krnn_smoother(X, y, k = 10)
plot_krnn_smoother(sm)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/classifiers.R
\name{bayes_depth_classify}
\alias{bayes_depth_classify}
\alias{print.fd_bayes_fit}
\title{Bayesian Kernel-Density Classifier on Signed Depth}
\usage{
bayes_depth_classify(
  X_train,
  y_train,
  X_test = NULL,
  priors = NULL,
  bw_method = "nrd0",
  grid = NULL,
  type = c("tukey", "simplicial")
)

\method{print}{fd_bayes_fit}(x, ...)
}
\arguments{
\item{X_train}{Numeric matrix (\eqn{n_{\mathrm{train}} \times M}).}

\item{y_train}{Integer vector of labels (0/1).}

\item{X_test}{Numeric matrix of test curves. If \code{NULL} the training
set is used.}

\item{priors}{Named numeric vector with elements \code{"0"} and \code{"1"}
giving prior probabilities. Defaults to empirical class proportions.}

\item{bw_method}{Bandwidth selection method passed to
\code{\link[stats]{density}}: \code{"nrd0"} (default), \code{"nrd"},
\code{"ucv"}, \code{"bcv"}, or \code{"SJ"}.}

\item{grid}{Numeric vector of length \eqn{M} (time grid).}

\item{type}{Depth variant.}

\item{x}{A \code{"fd_bayes_fit"} object.}

\item{...}{Ignored.}
}
\value{
An object of class \code{"fd_bayes_fit"} with components:
\describe{
\item{predicted}{Predicted class labels.}
\item{prob}{Matrix with columns \code{prob_0} and \code{prob_1}:
posterior probabilities.}
\item{log_odds}{Log-odds
\eqn{\log(f_0\pi_0 / f_1\pi_1)}.}
\item{sd_train}{The \code{"fd_signed_depth"} object for the training
set.}
}

Invisibly returns \code{x}.
}
\description{
Classifies functional curves via Bayes' rule applied to the signed depths.
Class-conditional densities \eqn{f_g(\mathrm{sdp})} are estimated by
kernel density estimation; prior probabilities are estimated from class
frequencies or supplied by the user.
}
\details{
The posterior probability that a new observation belongs to class \eqn{g}
is
\deqn{P(g \mid \mathrm{sdp}_0) =
  \frac{f_g(\mathrm{sdp}_0)\,\pi_g}
  {\sum_{g'} f_{g'}(\mathrm{sdp}_0)\,\pi_{g'}}.}
}
\examples{
set.seed(11)
M <- 80; N <- 100
t  <- seq(0, 1, length.out = M)
X0 <- t(replicate(50, sin(2 * pi * t) + rnorm(M, sd = 0.4)))
X1 <- t(replicate(50, cos(2 * pi * t) + rnorm(M, sd = 0.4)))
X  <- rbind(X0, X1)
y  <- rep(0:1, each = 50)
fit <- bayes_depth_classify(X, y)
table(fit$predicted, y)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/krnn.R
\name{krnn_fit}
\alias{krnn_fit}
\alias{print.krnn_fit}
\title{k-Ranked Nearest Neighbour Classifier for Functional Data}
\usage{
krnn_fit(
  X_train,
  y_train,
  X_test = NULL,
  k = 5L,
  grid = NULL,
  type = c("tukey", "simplicial"),
  sd_train = NULL
)

\method{print}{krnn_fit}(x, ...)
}
\arguments{
\item{X_train}{Numeric matrix (\eqn{n_{\mathrm{train}} \times M}) of
training curves (rows = curves, columns = time points).}

\item{y_train}{Integer vector of class labels (0/1) of length
\eqn{n_{\mathrm{train}}}.}

\item{X_test}{Numeric matrix of test curves. If \code{NULL} (default) the
training set is used for in-sample evaluation.}

\item{k}{Positive integer: half-neighbourhood size. The total number of
neighbours per observation is \eqn{2k}. Use \code{\link{krnn_cv}} to
select k.}

\item{grid}{Numeric vector of length \eqn{M} giving observation times.
Defaults to equally spaced points on \eqn{[0, 1]}.}

\item{type}{Depth variant passed to \code{\link{fm_depth}}: \code{"tukey"}
(default) or \code{"simplicial"}.}

\item{sd_train}{Optional pre-computed \code{"fd_signed_depth"} object for
the training set (avoids recomputation when called repeatedly).}

\item{x}{A \code{"krnn_fit"} object.}

\item{...}{Ignored.}
}
\value{
An object of class \code{"krnn_fit"} with components:
\describe{
\item{predicted}{Integer vector of predicted class labels.}
\item{prob}{Numeric vector of estimated probabilities
\eqn{\hat{P}(y=0|\mathrm{sdp})}.}
\item{sd_train}{The \code{"fd_signed_depth"} object for the training
set.}
\item{y_train}{The training labels.}
\item{k}{The value of k used.}
}

Invisibly returns \code{x}.
}
\description{
Fits the k-RNN classifier of Perez Ruiz (2020). Curves are ranked by their
signed depth and each new observation is assigned to the majority class
among its \eqn{2k} nearest ranked neighbours (k above and k below).
}
\examples{
set.seed(7)
M <- 100; n <- 80
t <- seq(0, 1, length.out = M)
X0 <- t(replicate(n / 2, sin(2 * pi * t) + rnorm(M, sd = 0.4)))
X1 <- t(replicate(n / 2, cos(2 * pi * t) + rnorm(M, sd = 0.4)))
X  <- rbind(X0, X1)
y  <- rep(0:1, each = n / 2)
idx_tr <- sample(n, 60)
fit <- krnn_fit(X[idx_tr, ], y[idx_tr], X[-idx_tr, ], k = 5)
print(fit)

}
\references{
Perez Ruiz, D. A. (2020). \emph{Supervised Classification for Functional
Data}. PhD thesis, University of Manchester. Section 3.2.
}
\seealso{
\code{\link{krnn_cv}}, \code{\link{predict.krnn_fit}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/krnn.R
\name{predict.krnn_fit}
\alias{predict.krnn_fit}
\title{Predict Method for krnn_fit Objects}
\usage{
\method{predict}{krnn_fit}(object, newdata, y_true = NULL, ...)
}
\arguments{
\item{object}{A \code{"krnn_fit"} object.}

\item{newdata}{Numeric matrix of new curves
(\eqn{n_{\mathrm{new}} \times M}).}

\item{y_true}{Optional integer vector of true labels for computing the
misclassification rate.}

\item{...}{Ignored.}
}
\value{
A list with components \code{predicted}, \code{prob}, and (if
\code{y_true} is supplied) \code{misclass}.
}
\description{
Classifies new functional observations using a fitted
\code{\link{krnn_fit}} object.
}
\examples{
set.seed(7)
M <- 100; n <- 80
t <- seq(0, 1, length.out = M)
X0 <- t(replicate(n / 2, sin(2 * pi * t) + rnorm(M, sd = 0.4)))
X1 <- t(replicate(n / 2, cos(2 * pi * t) + rnorm(M, sd = 0.4)))
X  <- rbind(X0, X1)
y  <- rep(0:1, each = n / 2)
idx_tr <- sample(n, 60)
fit  <- krnn_fit(X[idx_tr, ], y[idx_tr], k = 5)
pred <- predict(fit, newdata = X[-idx_tr, ], y_true = y[-idx_tr])
pred$misclass

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fm_depth.R
\name{fm_depth}
\alias{fm_depth}
\title{Fraiman-Muniz Functional Depth}
\usage{
fm_depth(X, grid = NULL, type = c("tukey", "simplicial"))
}
\arguments{
\item{X}{Numeric matrix of dimension \eqn{N \times M}, where rows are curves
and columns are time points.  Missing values are not allowed.}

\item{grid}{Numeric vector of length \eqn{M} giving the observation times
(e.g. \code{seq(0, 1, length.out = M)}).  If \code{NULL} (default) equally spaced
points on \eqn{[0, 1]} are assumed.}

\item{type}{Character string, either \code{"tukey"} (default) or \code{"simplicial"},
selecting the univariate depth used at each time point.}
}
\value{
A numeric vector of length \eqn{N} containing the FM depth value
for each curve.
}
\description{
Computes the Fraiman-Muniz (FM) depth for a collection of discretised
functional observations. Two variants are supported: the Tukey-FM depth,
which maps depth values to \eqn{[0, 1]}, and the Simplicial-FM depth,
which maps to \eqn{[0.5, 1]}.  The two are related by
\deqn{\text{Tukey-FM}_i = 2\,(\text{Simplicial-FM}_i - 1/2).}
}
\details{
For a discretised dataset \eqn{\{x_i(t_j)\}}, \eqn{i = 1,\ldots,N},
\eqn{j = 1,\ldots,M}, the Simplicial-FM depth is
\deqn{
  \text{FM}_i = \sum_{j=2}^{M}(t_j - t_{j-1})
  \left[1 - \left|\frac{1}{2} - F_{N,t_j}(x_i(t_j))\right|\right],
}
where \eqn{F_{N,t}} is the empirical CDF of the sample at time \eqn{t}.
The Tukey-FM depth substitutes the Tukey univariate depth
\eqn{\min\{F_{N,t}(x), 1 - F_{N,t}(x)\}} at each time point.
}
\examples{
set.seed(1)
N <- 50; M <- 100
X <- matrix(rnorm(N * M), nrow = N)
d <- fm_depth(X)
plot(d, xlab = "Curve index", ylab = "Tukey-FM depth")

}
\references{
Fraiman, R. and Muniz, G. (2001). Trimmed means for functional data.
\emph{Test}, \strong{10}(2), 419–440.

López-Pintado, S. and Romo, J. (2009). On the concept of depth for
functional data. \emph{Journal of the American Statistical Association},
\strong{104}(486), 718–734.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/krnn.R
\name{plot_krnn_smoother}
\alias{plot_krnn_smoother}
\title{Plot a krnn_smoother Object}
\usage{
plot_krnn_smoother(x, ...)
}
\arguments{
\item{x}{A \code{"krnn_smoother"} object returned by
\code{\link{krnn_smoother}}.}

\item{...}{Additional graphical parameters passed to
\code{\link[graphics]{plot.default}}.}
}
\value{
Invisibly returns \code{x}.
}
\description{
Plots the estimated conditional probability
\eqn{\hat{P}(y = 0 \mid \mathrm{sdp})} against the signed depth, with
optional bootstrap confidence bands.
}
\examples{
set.seed(5)
M <- 100; N <- 80
t  <- seq(0, 1, length.out = M)
X0 <- t(replicate(40, sin(2 * pi * t) + rnorm(M, sd = 0.4)))
X1 <- t(replicate(40, cos(2 * pi * t) + rnorm(M, sd = 0.4)))
X  <- rbind(X0, X1)
y  <- rep(0:1, each = 40)
sm <- krnn_smoother(X, y, k = 10)
plot_krnn_smoother(sm)

}
